"""Controller holding and managing HUE resources of type `light`."""
from __future__ import annotations

from typing import Optional, Tuple, Type

from ..models.feature import (
    AlertEffectType,
    AlertFeature,
    ColorFeature,
    ColorPoint,
    ColorTemperatureFeature,
    DimmingFeature,
    DynamicsFeature,
    OnFeature,
)
from ..models.light import Light
from ..models.resource import ResourceTypes
from .base import BaseResourcesController


class LightsController(BaseResourcesController[Type[Light]]):
    """Controller holding and managing HUE resources of type `light`."""

    item_type = ResourceTypes.LIGHT

    async def turn_on(self, id: str, transition_time: int | None = None) -> None:
        """Turn on the light."""
        await self.set_state(id, on=True, transition_time=transition_time)

    async def turn_off(self, id: str, transition_time: int | None = None) -> None:
        """Turn off the light."""
        await self.set_state(id, on=False, transition_time=transition_time)

    async def set_brightness(
        self, id: str, brightness: float, transition_time: int | None = None
    ) -> None:
        """Set brightness to light. Turn on light if it's currently off."""
        await self.set_state(
            id, on=True, brightness=brightness, transition_time=transition_time
        )

    async def set_color(
        self, id: str, x: float, y: float, transition_time: int | None = None
    ) -> None:
        """Set color to light. Turn on light if it's currently off."""
        await self.set_state(
            id, on=True, color_xy=(x, y), transition_time=transition_time
        )

    async def set_color_temperature(
        self, id: str, mirek: int, transition_time: int | None = None
    ) -> None:
        """Set Color Temperature to light. Turn on light if it's currently off."""
        await self.set_state(
            id, on=True, color_temp=mirek, transition_time=transition_time
        )

    async def set_state(
        self,
        id: str,
        on: bool = True,
        brightness: Optional[float] = None,
        color_xy: Optional[Tuple[float, float]] = None,
        color_temp: Optional[int] = None,
        transition_time: int | None = None,
        alert: AlertEffectType | None = None,
    ) -> None:
        """Set supported feature(s) to light resource."""
        update_obj = Light(id=id, on=OnFeature(on=on))
        if brightness is not None:
            update_obj.dimming = DimmingFeature(brightness=brightness)
        if color_xy is not None:
            update_obj.color = ColorFeature(xy=ColorPoint(*color_xy))
        if color_temp is not None:
            update_obj.color_temperature = ColorTemperatureFeature(mirek=color_temp)
        if transition_time is not None:
            update_obj.dynamics = DynamicsFeature(duration=transition_time)
        if alert is not None:
            update_obj.alert = AlertFeature(action=alert)
        await self.update(id, update_obj)



"""Controller holding and managing HUE resources of type `device`."""
from __future__ import annotations

from typing import List, Type

from ..models.connectivity import ZigbeeConnectivity
from ..models.device import Device
from ..models.light import Light
from ..models.resource import ResourceTypes
from ..models.room import Room
from .base import BaseResourcesController
from .sensors import SENSOR_TYPES


class DevicesController(BaseResourcesController[Type[Device]]):
    """Controller holding and managing HUE resources of type `device`."""

    item_type = ResourceTypes.DEVICE

    def get_lights(self, id: str) -> List[Light]:
        """Return all lights belonging to given device."""
        return [
            self._bridge.lights[x] for x in self[id].lights if x in self._bridge.lights
        ]

    def get_sensors(self, id: str) -> List[SENSOR_TYPES]:
        """Return all sensors belonging to given device."""
        return [
            self._bridge.sensors[x]
            for x in self[id].sensors
            if x in self._bridge.sensors
        ]

    def get_room(self, id: str) -> Room | None:
        """Return room the given device belongs to (if any)."""
        return next((x for x in self._bridge.groups.room if id in x.devices), None)

    def get_zigbee_connectivity(self, id: str) -> ZigbeeConnectivity | None:
        """Return the ZigbeeConnectivity resource connected to device."""
        for service in self._items[id].services:
            if service.rtype == ResourceTypes.ZIGBEE_CONNECTIVITY:
                return self._bridge.sensors.zigbee_connectivity[service.rid]
        return None



"""Model(s) for behavior_instance (automation) resource on HUE bridge."""
from dataclasses import dataclass

from .resource import ResourceTypes
from .script import ScriptInstance


@dataclass
class BehaviorInstance(ScriptInstance):
    """
    Represent BehaviorInstance object as used by the Hue api.

    clip-api.schema.json#/definitions/BehaviorInstanceGet
    clip-api.schema.json#/definitions/BehaviorInstancePost
    clip-api.schema.json#/definitions/BehaviorInstancePut
    """

    type: ResourceTypes = ResourceTypes.BEHAVIOR_INSTANCE



"""Model(s) for motion resource on HUE bridge."""
from dataclasses import dataclass


from typing import Optional

from .resource import SensingService, ResourceTypes


@dataclass
class MotionSensingFeature:
    """
    Represent MotionSensingFeature as retrieved from api.

    clip-api.schema.json#/definitions/MotionSensingFeatureGet
    """

    motion: Optional[bool] = None
    motion_valid: Optional[bool] = None


@dataclass
class Motion(SensingService):
    """
    Represent a Motion resource, a service detecting motion.

    clip-api.schema.json#/definitions/Motion
    """

    motion: Optional[MotionSensingFeature] = None
    type: ResourceTypes = ResourceTypes.MOTION



"""Model(s) for grouped_light resource on HUE bridge."""
from dataclasses import dataclass

from typing import Optional

from .feature import AlertFeature, OnFeatureBasic
from .resource import Resource, ResourceTypes


@dataclass
class GroupedLight(Resource):
    """
    Represent a GroupedLight object as used by the Hue api.

    clip-api.schema.json#/definitions/GroupedLightGet
    clip-api.schema.json#/definitions/GroupedLightPost
    clip-api.schema.json#/definitions/GroupedLightPut
    """

    on: Optional[OnFeatureBasic] = None
    alert: Optional[AlertFeature] = None
    type: ResourceTypes = ResourceTypes.GROUPED_LIGHT



"""Control a Philips Hue bridge with V2 API."""
from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
import logging
from types import TracebackType
from typing import Callable, Generator, List, Optional, Type

import aiohttp
from aiohttp import ClientResponse
from asyncio_throttle import Throttler
from aiohue.v2.models.clip import LOGGER, CLIPResource

from ..errors import Unauthorized, raise_from_error, BridgeBusy
from .controllers.events import EventCallBackType, EventStream, EventType

from .controllers.config import ConfigController
from .controllers.sensors import SensorsController
from .controllers.lights import LightsController
from .controllers.groups import GroupsController
from .controllers.scenes import ScenesController
from .controllers.devices import DevicesController

MAX_RETRIES = 25  # how many times do we retry on a 503 (bridge overload/rate limit)
THROTTLE_CONCURRENT_REQUESTS = 2  # how many concurrent requests to the bridge
THROTTLE_TIMESPAN = 0.25  # timespan/period (in seconds) for the rate limiting


class HueBridgeV2:
    """Control a Philips Hue bridge with V2 API."""

    def __init__(
        self,
        host: str,
        app_key: str,
        websession: aiohttp.ClientSession | None = None,
    ) -> None:
        """
        Initialize the Bridge instance.

        Parameters:
            `host`: the hostname or IP-address of the bridge as string.
            `app_key`: provide the hue appkey/username for authentication.
            `websession`: optionally provide a aiohttp ClientSession.
        """
        self._host = host
        self._app_key = app_key
        self._websession = websession
        self._websession_provided = websession is not None

        self.logger = logging.getLogger(f"{__package__}[{host}]")
        self._events = EventStream(self)
        # all resource controllers
        self._config = ConfigController(self)
        self._devices = DevicesController(self)
        self._lights = LightsController(self)
        self._scenes = ScenesController(self)
        self._groups = GroupsController(self)
        self._sensors = SensorsController(self)
        # Setup the Throttler/rate limiter for requests to the bridge.
        # NOTE: The EventStream also counts for 1 connection to the bridge.
        self._throttler = Throttler(
            rate_limit=THROTTLE_CONCURRENT_REQUESTS, period=THROTTLE_TIMESPAN
        )

    @property
    def bridge_id(self) -> str | None:
        """Return the ID of the bridge we're currently connected to."""
        return self._config.bridge_id

    @property
    def host(self) -> str:
        """Return the hostname of the bridge."""
        return self._host

    @property
    def events(self) -> EventStream:
        """Return the EventStream controller."""
        return self._events

    @property
    def config(self) -> ConfigController:
        """Get the Config Controller with config-like resources."""
        return self._config

    @property
    def devices(self) -> DevicesController:
        """Get the Devices Controller for managing all device resources."""
        return self._devices

    @property
    def lights(self) -> LightsController:
        """Get the Lights Controller for managing all light resources."""
        return self._lights

    @property
    def scenes(self) -> ScenesController:
        """Get the Scenes Controller for managing all scene resources."""
        return self._scenes

    @property
    def groups(self) -> GroupsController:
        """Get the Groups Controller for managing all group resources."""
        return self._groups

    @property
    def sensors(self) -> SensorsController:
        """Get the Sensors Controller for managing all sensor resources."""
        return self._sensors

    async def initialize(self) -> None:
        """Initialize the connection to the bridge and fetch all data."""
        # Initialize all HUE resource controllers
        # fetch complete full state once and distribute to controllers
        await self.fetch_full_state()
        # start event listener
        await self._events.initialize()
        # subscribe to reconnect event
        self._events.subscribe(self._handle_event, EventType.RECONNECTED)

    async def close(self) -> None:
        """Close connection and cleanup."""
        await self.events.stop()
        if not self._websession_provided:
            await self._websession.close()
        self.logger.info("Connection to bridge closed.")

    def subscribe(
        self,
        callback: EventCallBackType,
    ) -> Callable:
        """
        Subscribe to status changes for all resources.

        Returns:
            function to unsubscribe.
        """
        unsubscribes = [
            self.config.subscribe(callback),
            self.devices.subscribe(callback),
            self.groups.subscribe(callback),
            self.lights.subscribe(callback),
            self.scenes.subscribe(callback),
            self.sensors.subscribe(callback),
        ]

        def unsubscribe():
            for unsub in unsubscribes:
                unsub()

        return unsubscribe

    async def request(self, method: str, path: str, **kwargs) -> dict | List[dict]:
        """Make request on the api and return response data."""
        # The bridge will rate limit if we send more requests than about 2-5 per second
        # we guard ourselves from hitting the rate limit by using a throttler
        # but others apps/services are hitting the Hue bridge too so we still
        # might hit the rate limit/overload at some point so we also retry if this happens.
        retries = 0

        while retries < MAX_RETRIES:
            retries += 1

            if retries > 1:
                retry_wait = 0.25 * retries
                LOGGER.debug(
                    "Got 503 error from Hue bridge, retry request in %s seconds",
                    retry_wait,
                )
                await asyncio.sleep(retry_wait)

            async with self._throttler, self.create_request(
                method, path, **kwargs
            ) as resp:
                # 503 means the bridge is rate limiting/overloaded, we should back off a bit.
                if resp.status == 503:
                    continue
                if resp.status == 403:
                    raise Unauthorized
                # raise on all other error status codes
                resp.raise_for_status()
                result = await resp.json()
                if result.get("errors"):
                    raise_from_error(result["errors"][0])
                return result["data"]

        raise BridgeBusy(
            f"{retries} requests to the bridge failed, "
            "its probably overloaded. Giving up."
        )

    @asynccontextmanager
    async def create_request(
        self, method: str, path: str, **kwargs
    ) -> Generator[ClientResponse, None, None]:
        """
        Make a request to any path with V2 request method (auth in header).

        Returns a generator with aiohttp ClientResponse.
        """
        if self._websession is None:
            self._websession = aiohttp.ClientSession()

        url = f"https://{self._host}/{path}"

        kwargs["ssl"] = False

        if "headers" not in kwargs:
            kwargs["headers"] = {}

        kwargs["headers"]["hue-application-key"] = self._app_key

        async with self._websession.request(method, url, **kwargs) as res:
            yield res

    async def __aenter__(self) -> "HueBridgeV2":
        """Return Context manager."""
        await self.initialize()
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> Optional[bool]:
        """Exit context manager."""
        await self.close()
        if exc_val:
            raise exc_val
        return exc_type

    async def _handle_event(self, type: EventType, item: CLIPResource | None) -> None:
        """Handle incoming event for this resource from the EventStream."""
        if type != EventType.RECONNECTED:
            return
        await self.fetch_full_state()

    async def fetch_full_state(self) -> None:
        """Fetch state on all controllers."""
        full_state = await self.request("get", "clip/v2/resource")
        await asyncio.gather(
            self._config.initialize(full_state),
            self._devices.initialize(full_state),
            self._lights.initialize(full_state),
            self._scenes.initialize(full_state),
            self._sensors.initialize(full_state),
            self._groups.initialize(full_state),
        )



"""Controller holding and managing HUE resources that are of the config type."""
from __future__ import annotations

from typing import TYPE_CHECKING, Type, Union

from ...util import mac_from_bridge_id
from ..models.bridge import Bridge
from ..models.bridge_home import BridgeHome
from ..models.device import Device
from ..models.entertainment import Entertainment, EntertainmentConfiguration
from ..models.resource import ResourceTypes
from .base import BaseResourcesController, GroupedControllerBase

if TYPE_CHECKING:
    from .. import HueBridgeV2


class BridgeController(BaseResourcesController[Type[Bridge]]):
    """Controller holding and managing HUE resources of type `bridge`."""

    item_type = ResourceTypes.BRIDGE

    @property
    def id(self) -> str | None:
        """Return id of the only/first bridge found in items."""
        for item in self.items:
            return item.id
        return None


class BridgeHomeController(BaseResourcesController[Type[BridgeHome]]):
    """Controller holding and managing HUE resources of type `bridge_home`."""

    item_type = ResourceTypes.BRIDGE_HOME


class EntertainmentController(BaseResourcesController[Type[Entertainment]]):
    """Controller holding and managing HUE resources of type `entertainment`."""

    item_type = ResourceTypes.ENTERTAINMENT


class EntertainmentConfigurationController(
    BaseResourcesController[Type[EntertainmentConfiguration]]
):
    """Controller holding and managing HUE resources of type `entertainment_configuration`."""

    item_type = ResourceTypes.ENTERTAINMENT_CONFIGURATION


class ConfigController(
    GroupedControllerBase[
        Union[Bridge, BridgeHome, Entertainment, EntertainmentConfiguration]
    ]
):
    """
    Controller holding and managing HUE resources thare are of the config type.

    Note that the properties will raise AttributeError if not connected to a bridge.
    """

    @property
    def bridge_id(self) -> str:
        """Return bridge_id of bridge we're connected to."""
        return self.bridge.bridge_id

    @property
    def name(self) -> str:
        """Return name of bridge we're connected to."""
        return self.bridge_device.metadata.name

    @property
    def mac_address(self) -> str:
        """Return mac address of bridge we're connected to."""
        # the network mac is not available in api so we parse it from the id
        return mac_from_bridge_id(self.bridge_id)

    @property
    def model_id(self) -> str:
        """Return model ID of bridge we're connected to."""
        return self.bridge_device.product_data.model_id

    @property
    def software_version(self) -> str:
        """Return software version of bridge we're connected to."""
        return self.bridge_device.product_data.software_version

    @property
    def bridge(self) -> Bridge:
        """Return the only/first bridge found in items of resource `bridge`."""
        # the Hue resource system in V2 is generic and even the bridge object is returned as array
        # there should be only one object returned here
        return next((item for item in self.bridges))

    @property
    def bridge_device(self) -> Device:
        """Return the device object belonging to the bridge."""
        # the Hue resource system in V2 is generic and even the bridge metadata
        # can (only) be retrieved as a device object
        # do the plumbing by looking it up here.
        # this device object contains the metadata that is more or like comparable
        # with the V1 output like it's name and software version.
        if bridge := self.bridge:
            for device in self._bridge.devices:
                for service in device.services:
                    if service.rid == bridge.id:
                        return device
        raise AttributeError("bridge_device")

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize underlying controller instances."""
        self.bridges = BridgeController(bridge)
        self.bridge_home = BridgeHomeController(bridge)
        self.entertainment = EntertainmentController(bridge)
        self.entertainment_configuration = EntertainmentConfigurationController(bridge)
        super().__init__(
            bridge,
            [
                self.bridges,
                self.bridge_home,
                self.entertainment,
                self.entertainment_configuration,
            ],
        )



"""Model(s) for geofence_client resource on HUE bridge."""
from dataclasses import dataclass
from typing import Optional

from .resource import Resource, ResourceTypes


@dataclass
class GeofenceClient(Resource):
    """
    Representation of Geofence Client.

    clip-api.schema.json#/definitions/GeofenceClientGet
    clip-api.schema.json#/definitions/GeofenceClientPut
    """

    is_at_home: Optional[bool] = None  # Indicator if Geofence Client is at home.
    name: Optional[str] = None
    type: ResourceTypes = ResourceTypes.GEOFENCE_CLIENT


@dataclass
class GeofenceClientCreate:
    """
    Representation of Geofence Client.

    clip-api.schema.json#/definitions/GeofenceClientPost
    """

    is_at_home: Optional[bool] = None  # Indicator if Geofence Client is at home.
    name: Optional[str] = None
    type: ResourceTypes = ResourceTypes.GEOFENCE_CLIENT



"""Model(s) for temperature resource on HUE bridge."""

from dataclasses import dataclass


from typing import Optional

from .resource import SensingService, ResourceTypes


@dataclass
class TemperatureFeature:
    """Represent TemperatureFeature."""

    temperature: Optional[float]
    temperature_valid: Optional[bool]


@dataclass
class Temperature(SensingService):
    """
    Represent a Temperature resource, a sensor reporting Temperature.

    TODO: CLIP Schema missing for this resource.
    """

    temperature: Optional[TemperatureFeature] = None
    type: ResourceTypes = ResourceTypes.TEMPERATURE



"""Model(s) for Light resource on HUE bridge."""
from dataclasses import dataclass
from enum import Enum


from typing import Any, List, Optional, Type

from .feature import (
    AlertFeature,
    ColorFeature,
    ColorTemperatureFeature,
    DimmingFeature,
    DynamicsFeature,
    DynamicsFeatureStatus,
    OnFeature,
)
from .resource import (
    NamedResourceMetadata,
    Resource,
    ResourceTypes,
)


class LightArchetypes(Enum):
    """
    Enum with all possible LightArchetypes.

    clip-api.schema.json#/definitions/LightArchetypes
    """

    UNKNOWN = "unknown_archetype"
    CLASSIC_BULB = "classic_bulb"
    SULTAN_BULB = "sultan_bulb"
    FLOOD_BULB = "flood_bulb"
    SPOT_BULB = "spot_bulb"
    CANDLE_BULB = "candle_bulb"
    LUSTER_BULB = "luster_bulb"
    PENDANT_ROUND = "pendant_round"
    PENDANT_LONG = "pendant_long"
    CEILING_ROUND = "ceiling_round"
    CEILING_SQUARE = "ceiling_square"
    FLOOR_SHADE = "floor_shade"
    FLOOR_LANTERN = "floor_lantern"
    TABLE_SHADE = "table_shade"
    RECESSED_CEILING = "recessed_ceiling"
    RECESSED_FLOOR = "recessed_floor"
    SINGLE_SPOT = "single_spot"
    DOUBLE_SPOT = "double_spot"
    TABLE_WASH = "table_wash"
    WALL_LANTERN = "wall_lantern"
    WALL_SHADE = "wall_shade"
    FLEXIBLE_LAMP = "flexible_lamp"
    GROUND_SPOT = "ground_spot"
    WALL_SPOT = "wall_spot"
    PLUG = "plug"
    HUE_GO = "hue_go"
    HUE_LIGHTSTRIP = "hue_lightstrip"
    HUE_IRIS = "hue_iris"
    HUE_BLOOM = "hue_bloom"
    BOLLARD = "bollard"
    WALL_WASHER = "wall_washer"
    HUE_PLAY = "hue_play"
    VINTAGE_BULB = "vintage_bulb"
    CHRISTMAS_TREE = "christmas_tree"
    HUE_CENTRIS = "hue_centris"
    HUE_LIGHTSTRIP_TV = "hue_lightstrip_tv"

    @classmethod
    def _missing_(cls: Type, value: str):
        """Set default enum member if an unknown value is provided."""
        return LightArchetypes.UNKNOWN


@dataclass
class LightMetaData(NamedResourceMetadata):
    """
    Represent LightMetaData object as used by the Hue api.

    clip-api.schema.json#/definitions/LightMetaDataGet
    clip-api.schema.json#/definitions/LightMetaDataPut
    """

    archetype: Optional[LightArchetypes]


class LightModeValues(Enum):
    """
    Enum with all possible LightModes.

    clip-api.schema.json#/definitions/LightModeValues
    """

    NORMAL = "normal"
    STREAMING = "streaming"


@dataclass
class Gradient:
    """
    Represent Gradient object as used by the Hue api.

    TODO: Retrieve api spec.
    """

    points: Optional[List[Any]] = None
    points_capable: Optional[int] = None


@dataclass
class Light(Resource):
    """
    Represent a Light object as retrieved from the api.

    clip-api.schema.json#/definitions/LightGet
    clip-api.schema.json#/definitions/LightPut
    """

    metadata: Optional[LightMetaData] = None
    on: Optional[OnFeature] = None
    mode: Optional[LightModeValues] = None
    alert: Optional[AlertFeature] = None
    dimming: Optional[DimmingFeature] = None
    color_temperature: Optional[ColorTemperatureFeature] = None
    color: Optional[ColorFeature] = None
    dynamics: Optional[DynamicsFeature] = None
    gradient: Optional[Gradient] = None
    type: ResourceTypes = ResourceTypes.LIGHT

    @property
    def supports_dimming(self) -> bool:
        """Return if this light supports brightness control."""
        return self.dimming is not None

    @property
    def supports_color(self) -> bool:
        """Return if this light supports color control."""
        return self.color is not None

    @property
    def supports_color_temperature(self) -> bool:
        """Return if this light supports color_temperature control."""
        return self.color_temperature is not None

    @property
    def name(self) -> Optional[str]:
        """Return name from metadata."""
        if self.metadata is not None:
            return self.metadata.name
        return None

    @property
    def is_on(self) -> bool:
        """Return bool if light is currently powered on."""
        if self.on is not None:
            return self.on.on
        return False

    @property
    def brightness(self) -> float:
        """Return current brightness of light."""
        if self.dimming is not None:
            return self.dimming.brightness
        return 100.0 if self.is_on else 0.0

    @property
    def is_dynamic(self) -> bool:
        """Return bool if light is currently dynamically changing colors from a scene."""
        if self.dynamics is not None:
            return self.dynamics.status == DynamicsFeatureStatus.DYNAMIC_PALETTE
        return False

    @property
    def entertainment_active(self) -> bool:
        """Return bool if this light is currently streaming in Entertainment mode."""
        return self.mode == LightModeValues.STREAMING



"""Handle connecting to the HUE Eventstream and distribute events."""
from __future__ import annotations

import asyncio
import json
import random
import string
from asyncio.coroutines import iscoroutinefunction
from enum import Enum
from typing import TYPE_CHECKING, Callable, List, NoReturn, Tuple

from aiohttp import ClientTimeout
from aiohttp.client_exceptions import ClientError
from aiohue.v2.models.geofence_client import GeofenceClientCreate

if TYPE_CHECKING:
    from .. import HueBridgeV2

from ...errors import AiohueException, InvalidAPIVersion, InvalidEvent, Unauthorized
from ...util import NoneType
from ..models.clip import CLIPEvent, CLIPEventType, CLIPResource
from ..models.resource import ResourceTypes

CONNECTION_TIMEOUT = 90  # 90 seconds
KEEPALIVE_INTERVAL = 60  # every minute


class EventStreamStatus(Enum):
    """Status options of EventStream."""

    CONNECTING = 0
    CONNECTED = 1
    DISCONNECTED = 2


class EventType(Enum):
    """Enum with possible Events (based on CLIPEventType)."""

    # overriding Enum is not possible but we want the event names to match
    RESOURCE_ADDED = CLIPEventType.RESOURCE_ADDED.value
    RESOURCE_UPDATED = CLIPEventType.RESOURCE_UPDATED.value
    RESOURCE_DELETED = CLIPEventType.RESOURCE_DELETED.value
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    RECONNECTED = "reconnected"


EventCallBackType = Callable[[EventType, CLIPResource], None]
EventSubscriptionType = Tuple[
    EventCallBackType,
    "Tuple[EventType] | None",
    "Tuple[ResourceTypes] | None",
]


class EventStream:
    """Holds the connection to the HUE Clip EventStream."""

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize instance."""
        self._bridge = bridge
        self._listeners = set()
        self._event_queue = asyncio.Queue()
        self._last_event_id = ""
        self._status = EventStreamStatus.DISCONNECTED
        self._bg_tasks: List[asyncio.Task] = []
        self._subscribers: List[EventSubscriptionType] = []
        self._logger = bridge.logger.getChild("events")

    @property
    def connected(self) -> bool:
        """Return bool if we're connected."""
        return self._status == EventStreamStatus.CONNECTED

    @property
    def status(self) -> bool:
        """Return connection status."""
        return self._status

    async def initialize(self) -> None:
        """
        Start listening for events.

        Starts the connection to the Hue Eventstream and collect events.
        Connection will be auto-reconnected if it gets lost.
        """
        assert len(self._bg_tasks) == 0
        self._bg_tasks.append(asyncio.create_task(self.__event_reader()))
        self._bg_tasks.append(asyncio.create_task(self.__event_processor()))
        self._bg_tasks.append(asyncio.create_task(self.__keepalive_workaround()))

    async def stop(self) -> None:
        """Stop listening for events."""
        for task in self._bg_tasks:
            task.cancel()
        self._bg_tasks = []

    def subscribe(
        self,
        callback: Callable[[EventType, CLIPResource | None], None],
        event_filter: EventType | Tuple[EventType] | None = None,
        resource_filter: ResourceTypes | Tuple[ResourceTypes] | None = None,
    ) -> Callable:
        """
        Subscribe to events emitted by the Hue bridge for resources.

        Parameters:
            - `callback` - callback function to call when an event emits.
            - `event_filter` - Optionally provide an EventType as filter.
            - `resource_filter` - Optionally provide a ResourceType as filter.

        Returns:
            function to unsubscribe.
        """
        if not isinstance(event_filter, (NoneType, tuple)):
            event_filter = (event_filter,)
        if not isinstance(resource_filter, (NoneType, tuple)):
            resource_filter = (resource_filter,)
        subscription = (callback, event_filter, resource_filter)

        def unsubscribe():
            self._subscribers.remove(subscription)

        self._subscribers.append(subscription)
        return unsubscribe

    def emit(self, type: EventType, data: CLIPResource | None = None) -> None:
        """Emit event to all listeners."""
        for (callback, event_filter, resource_filter) in self._subscribers:
            if event_filter is not None and type not in event_filter:
                continue
            if (
                data is not None
                and resource_filter is not None
                and data.type not in resource_filter
            ):
                continue
            if iscoroutinefunction(callback):
                asyncio.create_task(callback(type, data))
            else:
                callback(type, data)

    async def __event_reader(self) -> NoReturn:
        """

        Read incoming SSE messages and put them in a Queue to be processed.

        Background task that keeps (re)connecting untill stopped.

        https://html.spec.whatwg.org/multipage/server-sent-events.html#event-stream-interpretation
        """
        self._status = EventStreamStatus.CONNECTING
        connect_attempts = 0
        headers = {
            "Accept": "text/event-stream",
            "Cache-Control": "no-cache",
            "Connection": "close",
        }

        while True:
            connect_attempts += 1
            # Messages come in line by line, according to EventStream/SSE specs
            # we iterate over the incoming lines in the streamreader.
            try:
                async with self._bridge.create_request(
                    "get",
                    "eventstream/clip/v2",
                    timeout=ClientTimeout(total=0, sock_read=CONNECTION_TIMEOUT),
                    headers=headers,
                ) as resp:
                    # update status to connected once we reach this point
                    self._status = EventStreamStatus.CONNECTED
                    if connect_attempts == 1:
                        self.emit(EventType.CONNECTED)
                    else:
                        self.emit(EventType.RECONNECTED)
                    connect_attempts = 1  # reset on succesfull connect
                    self._logger.debug("Connected to EventStream")
                    # read over incoming messages line by line
                    async for line in resp.content:
                        # process the message
                        self.__parse_message(line)
            except (ClientError, asyncio.TimeoutError) as err:
                # pass expected connection errors because we will auto retry
                status = getattr(err, "status", None)
                if status == 404:
                    raise InvalidAPIVersion from err
                if status == 403:
                    raise Unauthorized from err
                if status:
                    # for debugging purpose only
                    self._logger.debug(err)
            except Exception as err:
                # for debugging purpose only
                self._logger.exception(err)
                raise err

            # if we reach this point, the connection was lost
            self.emit(EventType.DISCONNECTED)

            reconnect_wait = min(2 * connect_attempts, 600)
            self._logger.debug(
                "Disconnected from EventStream"
                " - Reconnect will be attempted in %s seconds",
                reconnect_wait,
            )
            # every 10 failed connect attempts log warning
            if connect_attempts % 10 == 0:
                self._logger.warning(
                    "%s Attempts to (re)connect to bridge failed"
                    " - This might be an indication of connection issues.",
                    connect_attempts,
                )
            self._status = EventStreamStatus.CONNECTING
            await asyncio.sleep(reconnect_wait)

    async def __event_processor(self) -> NoReturn:
        """Process incoming CLIPEvents on the Queue and distribute those."""
        while True:
            event: CLIPEvent = await self._event_queue.get()
            # each clip event has array of updated/added/deleted objects in data property
            # we fire an event for each object that was added/updated/deleted
            for item in event.data:
                self.emit(EventType(event.type.value), item)

    def __parse_message(self, msg: bytes) -> None:
        """Parse a plain message string as received from EventStream."""
        try:
            line = msg.decode().strip()
            if not line or ":" not in line:
                return
            key, value = line.split(":", 1)
            if not key:
                return
            if key == "id":
                self._last_event_id = value.replace(":0", "")
                return
            if key == "data":
                # events is array with multiple events
                events: List[dict] = json.loads(value)
                for event in events:
                    clip_event = CLIPEvent.from_dict(event)
                    if clip_event.type == CLIPEventType.UNKNOWN:
                        raise InvalidEvent(
                            "Received invalid event %s", event.get("type")
                        )
                    self._event_queue.put_nowait(clip_event)
                return
            if key != "data":
                self._logger.debug("Received unexpected message: %s - %s", key, value)
        except Exception as exc:  # pylint: disable=broad-except
            self._logger.warning(
                "Unable to parse Event message: %s", line, exc_info=exc
            )

    async def __keepalive_workaround(self) -> NoReturn:
        """Send keepalive command to bridge, abusing geofence client."""
        # Oh yeah, this is a major hack and hopefully it will only be temporary ;-)
        # Signify forgot to implement some sort of periodic keep alive message on the EventBus
        # so we have no way to determine if the connection is still alive.
        # To workaround this, we create a geofence client (without a status)
        # on the bridge for aiohue which will have its name updated every minute
        # this will result in an event on the eventstream and thus a way to figure out
        # if its still alive. It's not very pretty but at least it works.
        # Now let's contact Signify if this can be solved.
        prefix = "aiohue_"

        while True:
            await asyncio.sleep(KEEPALIVE_INTERVAL)

            try:
                for client in self._bridge.sensors.geofence_client.items:
                    if client.name.startswith(prefix):
                        hass_client = client
                        break
                else:
                    await self._bridge.sensors.geofence_client.create(
                        GeofenceClientCreate(name=prefix)
                    )
                    continue

                # simply updating the name of the geofence client will result in an event message
                # the eventstream timeout will detect if our keepalive message was not received
                random_str = "".join(
                    (random.choice(string.ascii_lowercase)) for x in range(10)
                )
                hass_client.name = f"{prefix}{random_str}"
                await self._bridge.sensors.geofence_client.update(
                    hass_client.id, hass_client
                )
            except (ClientError, asyncio.TimeoutError, AiohueException) as err:
                # might happen on connection error, we don't want the retry logic to bail out
                self._logger.debug("Error while sending keepalive: %s", str(err))



"""Controller holding and managing HUE group resources."""
from __future__ import annotations

from typing import TYPE_CHECKING, List, Type, Union

from ..models.feature import OnFeature
from ..models.group import Group
from ..models.grouped_light import GroupedLight
from ..models.light import Light
from ..models.resource import ResourceTypes
from ..models.room import Room
from ..models.scene import Scene
from ..models.zone import Zone
from .base import BaseResourcesController, GroupedControllerBase

if TYPE_CHECKING:
    from .. import HueBridgeV2


class RoomController(BaseResourcesController[Type[Room]]):
    """Controller holding and managing HUE resources of type `room`."""

    item_type = ResourceTypes.ROOM

    def get_scenes(self, id: str) -> List[Scene]:
        """Get all scenes for this room."""
        return [scene for scene in self._bridge.scenes if scene.group.rid == id]


class ZoneController(BaseResourcesController[Type[Zone]]):
    """Controller holding and managing HUE resources of type `zone`."""

    item_type = ResourceTypes.ZONE

    def get_scenes(self, id: str) -> List[Scene]:
        """Get all scenes for this zone."""
        return [scene for scene in self._bridge.scenes if scene.group.rid == id]


class GroupedLightController(BaseResourcesController[Type[GroupedLight]]):
    """Controller holding and managing HUE resources of type `grouped_light`."""

    item_type = ResourceTypes.GROUPED_LIGHT

    def get_zone(self, id: str) -> Room | Zone | None:
        """Get the zone or room connected to grouped light."""
        for group in self._bridge.groups:
            if group.type == ResourceTypes.GROUPED_LIGHT:
                continue
            if group.grouped_light == id:
                return group
        return None

    def get_lights(self, id: str) -> List[Light]:
        """Return all underlying lights of this grouped light."""
        if zone := self.get_zone(id):
            return [self._bridge.lights[x] for x in zone.lights]
        return []  # fallback for a group without a zone (special 0 group)

    async def set_state(self, id: str, on: bool = True) -> None:
        """
        Set supported feature(s) to grouped_light resource.

        NOTE: a grouped_light can only handle OnFeature
        To send other features, you'll have to control the underlying lights
        """
        await self.update(id, GroupedLight(id, on=OnFeature(on=on)))


class GroupsController(GroupedControllerBase[Union[Room, Group, GroupedLight]]):
    """Controller grouping resources of both room and zone."""

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize instance."""
        self.grouped_light = GroupedLightController(bridge)
        self.room = RoomController(bridge)
        self.zone = ZoneController(bridge)
        super().__init__(
            bridge,
            [
                self.room,
                self.zone,
                self.grouped_light,
            ],
        )



"""Controller holding and managing HUE resources of type `scene`."""
from __future__ import annotations

from typing import Type

from aiohue.v2.models.room import Room
from aiohue.v2.models.zone import Zone

from ..models.feature import DimmingFeatureBasic, RecallAction, RecallFeature
from ..models.resource import ResourceTypes
from ..models.scene import Scene
from .base import BaseResourcesController


class ScenesController(BaseResourcesController[Type[Scene]]):
    """Controller holding and managing HUE resources of type `scene`."""

    item_type = ResourceTypes.SCENE

    async def recall(
        self,
        id: str,
        dynamic: bool = False,
        duration: int | None = None,
        brightness: float | None = None,
    ) -> None:
        """Turn on / recall scene."""
        action = RecallAction.DYNAMIC_PALETTE if dynamic else RecallAction.ACTIVE
        update_obj = Scene(
            id=id, recall=RecallFeature(action=action, duration=duration)
        )
        if brightness is not None:
            update_obj.recall.dimming = DimmingFeatureBasic(brightness=brightness)
        await self.update(id, update_obj)

    def get_group(self, id: str) -> Zone | Room:
        """Get group attached to given scene id."""
        scene = self[id]
        return next((x for x in self._bridge.groups if x.id == scene.group.rid))



"""Controller holding and managing HUE resources of sensor type."""
from __future__ import annotations

from typing import TYPE_CHECKING, Type, Union

from ..models.connectivity import ZigbeeConnectivity
from ..models.device_power import DevicePower
from ..models.geofence_client import GeofenceClient
from ..models.light_level import LightLevel
from ..models.temperature import Temperature

from ..models.button import Button
from ..models.motion import Motion
from ..models.resource import ResourceTypes
from .base import BaseResourcesController, GroupedControllerBase

if TYPE_CHECKING:
    from .. import HueBridgeV2

SENSOR_TYPES = Union[
    DevicePower,
    Button,
    GeofenceClient,
    LightLevel,
    Motion,
    Temperature,
    ZigbeeConnectivity,
]


class DevicePowerController(BaseResourcesController[Type[DevicePower]]):
    """Controller holding and managing HUE resources of type `device_power`."""

    item_type = ResourceTypes.DEVICE_POWER


class ButtonController(BaseResourcesController[Type[Button]]):
    """Controller holding and managing HUE resources of type `button`."""

    item_type = ResourceTypes.BUTTON


class GeofenceClientController(BaseResourcesController[Type[GeofenceClient]]):
    """Controller holding and managing HUE resources of type `geofence_client`."""

    item_type = ResourceTypes.GEOFENCE_CLIENT


class LightLevelController(BaseResourcesController[Type[LightLevel]]):
    """Controller holding and managing HUE resources of type `light_level`."""

    item_type = ResourceTypes.LIGHT_LEVEL

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, LightLevel(id=id, enabled=enabled))


class MotionController(BaseResourcesController[Type[Motion]]):
    """Controller holding and managing HUE resources of type `motion`."""

    item_type = ResourceTypes.MOTION

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, Motion(id=id, enabled=enabled))


class TemperatureController(BaseResourcesController[Type[Temperature]]):
    """Controller holding and managing HUE resources of type `temperature`."""

    item_type = ResourceTypes.TEMPERATURE


class ZigbeeConnectivityController(BaseResourcesController[Type[ZigbeeConnectivity]]):
    """Controller holding and managing HUE resources of type `zigbee_connectivity`."""

    item_type = ResourceTypes.ZIGBEE_CONNECTIVITY


class SensorsController(GroupedControllerBase[SENSOR_TYPES]):
    """Controller grouping resources of all sensor resources."""

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize instance."""
        self.button = ButtonController(bridge)
        self.device_power = DevicePowerController(bridge)
        self.geofence_client = GeofenceClientController(bridge)
        self.light_level = LightLevelController(bridge)
        self.motion = MotionController(bridge)
        self.temperature = TemperatureController(bridge)
        self.zigbee_connectivity = ZigbeeConnectivityController(bridge)
        super().__init__(
            bridge,
            [
                self.button,
                self.device_power,
                self.geofence_client,
                self.light_level,
                self.motion,
                self.temperature,
                self.zigbee_connectivity,
            ],
        )



"""Model(s) for entertainment resource on HUE bridge."""
from dataclasses import dataclass
from enum import Enum


from typing import List, Optional

from .feature import Position
from .resource import NamedResourceMetadata, Resource, ResourceIdentifier, ResourceTypes


@dataclass
class Segment:
    """
    Represent a Segment object as used by the Hue api.

    Contains the segments configuration of the device for entertainment purposes.
    A device can be segmented in a single way.

    clip-api.schema.json#/definitions/Segment
    """

    length: int
    start: int


@dataclass
class SegmentReference:
    """
    Represent a SegmentReference object as used by the Hue api.

    From which entertainment_service, which segment is a member of this channel.

    clip-api.schema.json#/definitions/SegmentReferenceGet
    """

    service: ResourceIdentifier
    index: int


@dataclass
class SegmentationProperties:
    """
    Represent a SegmentationProperties dict type.

    All properties regarding the segment capabilities of this device:
    the configuratibility, max_segments and all segment tables.
    clip-api.schema.json#/definitions/Segment
    """

    configurable: bool
    max_segments: int
    segments: List[Segment]


@dataclass
class EntertainmentChannelGet:
    """
    Represent a EntertainmentChannel object as used by the Hue api.

    Holds position of the channel and list of members, which is an array of segments resource identifiers.

    clip-api.schema.json#/definitions/EntertainmentChannelGet
    """

    # Bridge assigns a number (0...255) upon creation.
    # This is the number to be used by the HueStream API when addressing the channel
    channel_id: int
    # xyz position of this channel. It is the average position of its members.
    position: Position
    members: List[SegmentReference]


class EntertainmentConfigurationType(Enum):
    """
    Enum with possible Entertainment Configuration Types.

    as defined in:
    clip-api.schema.json#/definitions/EntertainmentConfiguration
    """

    SCREEN = "screen"  # Channels are organized around content from a screen
    MUSIC = "music"  # Channels are organized for music synchronization
    THREEDEESPACE = "3dspace"  # Channels are organized to provide 3d spacial effects
    OTHER = "other"  # General use case "


class EntertainmentStatus(Enum):
    """
    Enum with possible Entertainment Statuses.

    as defined in:
    clip-api.schema.json#/definitions/EntertainmentConfiguration
    """

    ACTIVE = "active"
    INACTIVE = "inactive"


class StreamingProxyMode(Enum):
    """
    Enum with possible StreamingProxy Modes.

    as defined in:
    clip-api.schema.json#/definitions/StreamingProxy
    """

    AUTO = "auto"
    MANUAL = "manual"


@dataclass
class StreamingProxy:
    """
    Represent a StreamingProxy object as used by the Hue api.

    Holds position of the channel and list of members, which is an array of segments resource identifiers.

    clip-api.schema.json#/definitions/StreamingProxyGet
    """

    mode: StreamingProxyMode
    node: ResourceIdentifier


@dataclass
class ServiceLocation:
    """
    Represent a ServiceLocation object as used by the Hue api.

    clip-api.schema.json#/definitions/ServiceLocationGet
    """

    service: ResourceIdentifier
    position: Position
    positions: Optional[List[Position]] = None


@dataclass
class EntertainmentLocations:
    """
    Represent a EntertainmentLocations object as used by the Hue api.

    The position of each entertainment service in the configuration.
    If the service has several segments, more than one x,y,z location may be needed

    clip-api.schema.json#/definitions/EntertainmentLocationsGet
    """

    service_locations: List[ServiceLocation]


class EntertainmentConfigurationAction(Enum):
    """
    Enum with possible Entertainment Configuration Actions.

    as defined in:
    clip-api.schema.json#/definitions/EntertainmentConfigurationPut
    """

    START = "start"
    STOP = "stop"


@dataclass
class EntertainmentConfiguration(Resource):
    """
    Entertainment Configuration as used by the Hue api.

    clip-api.schema.json#/definitions/EntertainmentConfigurationGet
    clip-api.schema.json#/definitions/EntertainmentConfigurationPut
    """

    # Friendly name of the entertainment configuration (max 32 chars)
    name: Optional[str] = None
    # Defines for which type of application this channel assignment was optimized
    configuration_type: Optional[EntertainmentConfigurationType] = None
    # Read only field reporting if the stream is active or not
    status: Optional[EntertainmentStatus] = None
    # who's streaming: Expected value is of a ResourceIdentifier of the type auth_v1
    active_streamer: Optional[ResourceIdentifier] = None
    # the proxy that is in use
    stream_proxy: Optional[StreamingProxy] = None
    # Holds the channels. Each channel groups segments of one or different light
    channels: Optional[List[EntertainmentChannelGet]] = None
    # Holds the lights connected to this entertainment setup
    light_services: Optional[List[ResourceIdentifier]] = None
    # Entertertainment services of the lights that are in the zone have locations
    locations: Optional[EntertainmentLocations] = None
    # action: only available on update/put
    # If status is "inactive" -> write start to start streaming.
    # Writing start when it's already active does not change the owership of the streaming.
    # If status is "active" -> write "stop" to end the current streaming.
    # In order to start streaming when other application is already streaming,
    # first write "stop" and then "start"
    action: Optional[EntertainmentConfigurationAction] = None
    metadata: Optional[NamedResourceMetadata] = None
    type: Optional[ResourceTypes] = ResourceTypes.ENTERTAINMENT


@dataclass
class Entertainment(Resource):
    """
    Represent Entertainment resource as used by the Hue api.

    clip-api.schema.json#/definitions/EntertainmentGet
    """

    renderer: Optional[bool] = None
    proxy: Optional[bool] = None
    segments: Optional[SegmentationProperties] = None
    type: ResourceTypes = ResourceTypes.ENTERTAINMENT



"""Model(s) for button resource on HUE bridge."""
from dataclasses import dataclass
from enum import Enum


from typing import Optional, Type

from .resource import Resource, ResourceTypes


class ButtonEvent(Enum):
    """
    Enum with possible button events.

    clip-api.schema.json#/definitions/ButtonEvent
    """

    INITIAL_PRESS = "initial_press"
    REPEAT = "repeat"
    SHORT_RELEASE = "short_release"
    LONG_RELEASE = "long_release"
    DOUBLE_SHORT_RELEASE = "double_short_release"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: str):
        """Set default enum member if an unknown value is provided."""
        return ButtonEvent.UNKNOWN


@dataclass
class ButtonFeature:
    """
    Represent ButtonFeature object as used by the Hue api.

    clip-api.schema.json#/definitions/ButtonFeature
    """

    last_event: ButtonEvent


@dataclass
class SwitchInputMetadata:
    """
    Represent SwitchInputMetadata object as used by the Hue api.

    clip-api.schema.json#/definitions/SwitchInputMetadata
    """

    # number of control within switch (value between 0..8).
    # Meaning in combination with type
    # - dots: Number of dots
    # - number: Number printed on device
    # - other: a logical order of controls in switch
    control_id: int


@dataclass
class Button(Resource):
    """
    Represent Button object as used by the Hue api.

    clip-api.schema.json#/definitions/Button
    """

    metadata: Optional[SwitchInputMetadata] = None
    button: Optional[ButtonFeature] = None
    type: ResourceTypes = ResourceTypes.BUTTON



"""Model(s) for Zone resource on HUE bridge."""
from dataclasses import dataclass


from typing import Optional

from .group import Group
from .resource import ResourceTypes
from .room import RoomMetadata


@dataclass
class Zone(Group):
    """
    Represent Zone object as used by the Hue api.

    A group grouping only services.
    A service can be in an arbitrary amount of groups.
    Following services are allowed:
        - light - relative_rotary - temperature - lightlevel - motion

    clip-api.schema.json#/definitions/Zone
    """

    metadata: Optional[RoomMetadata] = None
    type: ResourceTypes = ResourceTypes.ZONE

    @property
    def grouped_light(self) -> Optional[str]:
        """Return the grouped light id that is connected to this zone (if any)."""
        if not self.grouped_services:
            return None
        return next(
            (
                x.rid
                for x in self.grouped_services
                if x.rtype == ResourceTypes.GROUPED_LIGHT
            ),
            None,
        )



"""Model(s) for Scene resource on HUE bridge."""
from dataclasses import dataclass


from typing import List, Optional

from .feature import (
    ColorFeatureBasic,
    ColorTemperatureFeatureBasic,
    DimmingFeatureBasic,
    OnFeatureBasic,
    PaletteFeature,
    RecallFeature,
)
from .resource import (
    Resource,
    ResourceIdentifier,
    ResourceTypes,
)


@dataclass
class ActionAction:
    """
    Represent ActionAction object as used by the Hue api.

    The action to be executed on recall.
    clip-api.schema.json#/definitions/ActionAction
    clip-api.schema.json#/definitions/ActionActionPost
    clip-api.schema.json#/definitions/ActionActionPut
    """

    on: Optional[OnFeatureBasic] = None
    dimming: Optional[DimmingFeatureBasic] = None
    color: Optional[ColorFeatureBasic] = None
    color_temperature: Optional[ColorTemperatureFeatureBasic] = None


@dataclass
class Action:
    """
    Represent Action object as used by the Hue api.

    clip-api.schema.json#/definitions/ActionGet
    clip-api.schema.json#/definitions/ActionPut
    clip-api.schema.json#/definitions/ActionPost
    """

    target: ResourceIdentifier
    action: ActionAction


@dataclass
class SceneMetadata:
    """
    Represent SceneMetadata object as used by the Hue api.

    clip-api.schema.json#/definitions/SceneMetadataGet
    clip-api.schema.json#/definitions/SceneMetadataPost
    clip-api.schema.json#/definitions/SceneMetadataPut
    """

    name: str
    image: Optional[ResourceIdentifier] = None


@dataclass
class SceneService(Resource):
    """
    Represent SceneService object as used by the Hue api.

    clip-api.schema.json#/definitions/SceneServiceGet
    clip-api.schema.json#/definitions/SceneServicePost
    clip-api.schema.json#/definitions/SceneServicePut
    """

    actions: Optional[List[Action]] = None
    recall: Optional[RecallFeature] = None  # used only on update/set


@dataclass
class Scene(SceneService):
    """
    Represent Scene object as used by the Hue api.

    Inherited from SceneService.
    clip-api.schema.json#/definitions/SceneGet
    clip-api.schema.json#/definitions/ScenePost
    clip-api.schema.json#/definitions/ScenePut
    """

    group: Optional[ResourceIdentifier] = None
    metadata: Optional[SceneMetadata] = None
    # speed: current transition speed (readonly)
    speed: Optional[float] = None
    palette: Optional[PaletteFeature] = None
    type: ResourceTypes = ResourceTypes.SCENE



"""Model(s) for bridge resource on HUE bridge."""
from dataclasses import dataclass


from typing import Optional

from .resource import Resource, ResourceTypes


@dataclass
class TimeZone:
    """Represent TimeZone object as received from API."""

    time_zone: str  # e.g. Europe/Amsterdam


@dataclass
class Bridge(Resource):
    """
    Represent Bridge object as retrieved from the api.

    clip-api.schema.json#/definitions/BridgeGet
    """

    bridge_id: Optional[str] = None  # sent on get, can not be updated
    time_zone: Optional[TimeZone] = None
    type: ResourceTypes = ResourceTypes.BRIDGE



"""Utils for aiohue."""
from __future__ import annotations
from enum import Enum
import logging
from dataclasses import asdict, fields, is_dataclass, dataclass
from typing import Any, Type, Union, get_args, get_origin
from aiohttp import ClientSession
from datetime import datetime

from aiohue.errors import raise_from_error

try:
    # python 3.10
    from types import NoneType
except:  # noqa
    # older python version
    NoneType = type(None)


async def create_app_key(
    host: str, device_type: str, websession: ClientSession | None = None
) -> str:
    """
    Create a user on the Hue bridge and return it's app_key for authentication.

    The link button on the bridge must be pressed before executing this call,
    otherwise a LinkButtonNotPressed error will be raised.

    Parameters:
        `host`: the hostname or IP-address of the bridge as string.
        `device_type`: provide a name/type for your app for identification.
        `websession`: optionally provide a aiohttp ClientSession.
    """
    # https://developers.meethue.com/documentation/configuration-api#71_create_user
    # there is not (yet) a V2 way of creating a user.
    # so this can be used for both V1 and V2 bridges (for now).
    data = {"devicetype": device_type}
    websession_provided = websession is not None
    if websession is None:
        websession = ClientSession()
    try:
        # try both https and http
        for proto in ["https", "http"]:
            try:
                url = f"{proto}://{host}/api"
                async with websession.post(url, json=data, ssl=False) as resp:
                    resp.raise_for_status()
                    result = await resp.json()
                    # response is returned as list
                    result = result[0]
                    if "error" in result:
                        raise_from_error(result["error"])
                    return result["success"]["username"]
            except Exception as exc:  # pylint: disable=broad-except
                if proto == "http":
                    raise exc
    finally:
        if not websession_provided:
            await websession.close()


def normalize_bridge_id(bridge_id: str):
    """Normalize a bridge identifier."""
    bridge_id = bridge_id.lower()

    # zeroconf: properties['id'], field contains semicolons after each 2 char
    if len(bridge_id) == 17 and bridge_id.count(":") == 5:
        return bridge_id.replace(":", "")

    # nupnp: contains 4 extra characters in the middle: "fffe"
    if len(bridge_id) == 16 and bridge_id[6:10] == "fffe":
        return bridge_id[0:6] + bridge_id[-6:]

    # SSDP/UPNP and Hue Bridge API contains right ID.
    if len(bridge_id) == 12:
        return bridge_id

    logging.getLogger(__name__).warn("Received unexpected bridge id: %s", bridge_id)

    return bridge_id


def update_dataclass(org_obj: dataclass, new_obj: dataclass):
    """Update instance of dataclass with another, skipping None values."""
    for f in fields(new_obj):
        new_val = getattr(new_obj, f.name, None)
        cur_val = getattr(org_obj, f.name, None)
        if new_val is None:
            continue
        if cur_val == new_val:
            continue
        if is_dataclass(new_val) and cur_val is not None:
            # the None guard is in place for nested structures with optional data
            update_dataclass(getattr(org_obj, f.name), new_val)
        else:
            setattr(org_obj, f.name, new_val)


def dataclass_to_dict(obj_in: dataclass, skip_none: bool = True) -> dict:
    """Convert dataclass instance to dict, optionally skip None values."""
    if skip_none:
        dict_obj = asdict(
            obj_in, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
        )
    else:
        dict_obj = asdict(obj_in)

    def _clean_dict(_dict_obj: dict):
        final = {}
        for key, value in _dict_obj.items():
            if value is None and skip_none:
                continue
            if isinstance(value, dict):
                value = _clean_dict(value)
            if isinstance(value, Enum):
                value = value.value
            final[key] = value
        return final

    return _clean_dict(dict_obj)


def parse_utc_timestamp(datetimestr: str):
    """Parse datetime from string."""
    return datetime.fromisoformat(datetimestr.replace("Z", "+00:00"))


def dataclass_from_dict(cls: dataclass, dict_obj: dict, strict=False):
    """
    Create (instance of) a dataclass by providing a dict with values.

    Including support for nested structures and common type conversions.
    If strict mode enabled, any additional keys in the provided dict will result in a KeyError.
    """
    if strict:
        extra_keys = dict_obj.keys() - set([f.name for f in fields(cls)])
        if extra_keys:
            raise KeyError(
                "Extra key(s) %s not allowed for %s" % ",".join(extra_keys), (str(cls))
            )

    def _get_val(name: str, value: Any, value_type: Type):
        if value is None and value_type is NoneType:
            return None
        if is_dataclass(value_type) and isinstance(value, dict):
            return dataclass_from_dict(value_type, value)
        origin = get_origin(value_type)
        if origin is list:
            return [
                _get_val(name, subval, get_args(value_type)[0])
                for subval in value
                if subval is not None
            ]
        if origin is Union:  # type: ignore
            # try all possible types
            sub_value_types = get_args(value_type)
            for sub_arg_type in sub_value_types:
                if value is NoneType and sub_arg_type is NoneType:
                    return value
                # try them all until one succeeds
                try:
                    return _get_val(name, value, sub_arg_type)
                except (KeyError, TypeError, ValueError):
                    pass
            # if we get to this point, all possibilities failed
            # find out if we should raise or log this
            err = (
                f"Value {value} of type {type(value)} is invalid for {name}, "
                f"expected value of type {value_type}"
            )
            if NoneType not in sub_value_types:
                # raise exception, we have no idea how to handle this value
                raise TypeError(err)
            # failed to parse the (sub) value but None allowed, log only
            logging.getLogger(__name__).warn(err)
            return None
        if value_type is Any:
            return value
        if value is None and value_type is not NoneType:
            raise KeyError(f"`{name}` of type `{value_type}` is required.")
        if issubclass(value_type, Enum):
            return value_type(value)
        if value_type is type(datetime):
            return parse_utc_timestamp(value)
        if not isinstance(value, value_type):
            raise TypeError(
                f"Value {value} of type {type(value)} is invalid for {name}, "
                f"expected value of type {value_type}"
            )
        return value_type(value)

    return cls(
        **{
            field.name: _get_val(
                f"{cls.__name__}.{field.name}", dict_obj.get(field.name), field.type
            )
            for field in fields(cls)
        }
    )


def mac_from_bridge_id(bridge_id: str) -> str:
    """Parse mac address from bridge id."""
    parts = [
        bridge_id[0:2],
        bridge_id[2:4],
        bridge_id[4:6],
        # part 6:10 needs to be left out (fffe)
        bridge_id[10:12],
        bridge_id[12:14],
        bridge_id[14:16],
    ]
    return ":".join(parts)



"""Model(s) for Room resource on HUE bridge."""
from dataclasses import dataclass
from enum import Enum


from typing import Optional, Set, Type

from .group import Group
from .resource import (
    NamedResourceMetadata,
    ResourceTypes,
)


class RoomArchetype(Enum):
    """Possible archetypes of a room."""

    LIVING_ROOM = "living_room"
    KITCHEN = "kitchen"
    DINING = "dining"
    BEDROOM = "bedroom"
    KIDS_BEDROOM = "kids_bedroom"
    BATHROOM = "bathroom"
    NURSERY = "nursery"
    RECREATION = "recreation"
    OFFICE = "office"
    GYM = "gym"
    HALLWAY = "hallway"
    TOILET = "toilet"
    FRONT_DOOR = "front_door"
    GARAGE = "garage"
    TERRACE = "terrace"
    GARDEN = "garden"
    DRIVEWAY = "driveway"
    CARPORT = "carport"
    HOME = "home"
    DOWNSTAIRS = "downstairs"
    UPSTAIRS = "upstairs"
    TOP_FLOOR = "top_floor"
    ATTIC = "attic"
    GUEST_ROOM = "guest_room"
    STAIRCASE = "staircase"
    LOUNGE = "lounge"
    MAN_CAVE = "man_cave"
    COMPUTER = "computer"
    STUDIO = "studio"
    MUSIC = "music"
    TV = "tv"
    READING = "reading"
    CLOSET = "closet"
    STORAGE = "storage"
    LAUNDRY_ROOM = "laundry_room"
    BALCONY = "balcony"
    PORCH = "porch"
    BARBECUE = "barbecue"
    POOL = "pool"
    OTHER = "other"

    @classmethod
    def _missing_(cls: Type, value: str):
        """Set default enum member if an unknown value is provided."""
        return RoomArchetype.OTHER


@dataclass
class RoomMetadata(NamedResourceMetadata):
    """
    Represent RoomMetadata object as used by the Hue api.

    Configuration object for a room.

    clip-api.schema.json#/definitions/RoomMetadataGet
    clip-api.schema.json#/definitions/RoomMetadataPost
    clip-api.schema.json#/definitions/RoomMetadataPut
    """

    archetype: Optional[RoomArchetype] = None  # required in post/get, optional in put


@dataclass
class Room(Group):
    """
    Represent Room object as used by the Hue api.

    Room resources groups all devices within the same physical space of a room.
    Room only allows resources of type "device" as childs.
    A device can only be in at most one room.

    clip-api.schema.json#/definitions/RoomGet
    clip-api.schema.json#/definitions/RoomPost
    clip-api.schema.json#/definitions/RoomPut
    """

    metadata: Optional[RoomMetadata] = None  # required in post/get, optional in put
    type: ResourceTypes = ResourceTypes.ROOM

    @property
    def devices(self) -> Set[str]:
        """Return set of device id's that belong to this room."""
        return {x.rid for x in self.services}

    @property
    def grouped_light(self) -> Optional[str]:
        """Return the grouped light id that is connected to this room (if any)."""
        if not self.grouped_services:
            return None
        return next(
            (
                x.rid
                for x in self.grouped_services
                if x.rtype == ResourceTypes.GROUPED_LIGHT
            ),
            None,
        )



"""Feature Schemas used by various Hue resources."""

from dataclasses import dataclass, field
from enum import Enum


from typing import List, Optional, Type


@dataclass
class OnFeatureBasic:
    """
    Represent OnFeatureBasic object as used by the Hue api.

    clip-api.schema.json#/definitions/OnFeatureBasic
    """

    on: bool


@dataclass
class OnFeature(OnFeatureBasic):
    """
    Represent OnFeature object, inherited from OnFeatureBasic.

    clip-api.schema.json#/definitions/OnFeature
    """


@dataclass
class DimmingFeatureBasic:
    """
    Represent DimmingFeatureBasic object as used by the Hue api.

    clip-api.schema.json#/definitions/DimmingFeatureBasic
    """

    # Brightness percentage between 0 and 100.
    brightness: float


@dataclass
class DimmingFeature(DimmingFeatureBasic):
    """
    Represent DimmingFeature object, inherits from DimmingFeatureBasic.

    clip-api.schema.json#/definitions/DimmingFeature
    """


@dataclass
class Position:
    """
    Represent Position object as used by the Hue api.

    Each Coordinate value is a float between -1 and 1.
    clip-api.schema.json#/definitions/Position
    """

    x: float
    y: float
    z: float


class AlertEffectType(Enum):
    """Enum with possible alert effect values."""

    BREATHE = "breathe"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: str):
        """Set default enum member if an unknown value is provided."""
        return AlertEffectType.UNKNOWN


@dataclass
class AlertFeature:
    """Represent AlertFeature object as used by the Hue api."""

    action: Optional[AlertEffectType] = None
    action_values: List[AlertEffectType] = field(default_factory=list)


@dataclass
class ColorPoint:
    """
    CIE XY gamut position.

    Each position value is a float between -1 and 1.
    clip-api.schema.json#/definitions/ColorPoint
    """

    x: float
    y: float


@dataclass
class ColorFeatureGamut:
    """
    Color gamut of color bulb.

    Some bulbs do not properly return the Gamut information. In this case this is not present.
    clip-api.schema.json#/definitions/ColorFeatureGamut
    """

    red: ColorPoint
    green: ColorPoint
    blue: ColorPoint


class GamutType(Enum):
    """Enum with possible Gamut types."""

    A = "A"
    B = "B"
    C = "C"
    MIXED = "mixed"
    OTHER = "other"


@dataclass
class ColorFeatureBasic:
    """
    Represent ColorFeatureBasic object as used by the Hue api.

    clip-api.schema.json#/definitions/ColorFeatureBasic
    """

    xy: ColorPoint


@dataclass
class ColorFeature(ColorFeatureBasic):
    """
    Represent ColorFeature object as used by the Hue api.

    inherited from ColorFeatureBasic
    clip-api.schema.json#/definitions/ColorFeature
    """

    gamut_type: Optional[GamutType] = None
    gamut: Optional[ColorFeatureGamut] = None


@dataclass
class MirekSchema:
    """Represent Mirek schema."""

    mirek_minimum: int = 153
    mirek_maximum: int = 500

    def __post_init__(self):
        """Auto correct invalid values."""
        # Fix for devices that provide wrong info
        if self.mirek_minimum == 0:
            self.mirek_minimum = 153
        if self.mirek_maximum == 65535:
            self.mirek_maximum = 500


@dataclass
class ColorTemperatureFeatureBasic:
    """
    Represent ColorTemperatureFeatureBasic object as used by the Hue api.

    clip-api.schema.json#/definitions/ColorTemperatureFeatureBasic
    """

    # Color temperature in mirek (153-500) or None when the light color is not in the ct spectrum.
    mirek: Optional[int] = None
    mirek_schema: Optional[MirekSchema] = None
    # mirek_valid will be false if light is currently not in the ct spectrum
    mirek_valid: Optional[bool] = False

    def __post_init__(self):
        """Auto correct invalid values."""
        if self.mirek is None or self.mirek_schema is None:
            return
        if self.mirek < self.mirek_schema.mirek_minimum:
            self.mirek = self.mirek_schema.mirek_minimum
        if self.mirek > self.mirek_schema.mirek_maximum:
            self.mirek = self.mirek_schema.mirek_maximum


@dataclass
class ColorTemperatureFeature(ColorTemperatureFeatureBasic):
    """
    Represent ColorTemperatureFeature object, inherited from ColorTemperatureFeatureBasic.

    clip-api.schema.json#/definitions/ColorTemperatureFeature
    """


class DynamicsFeatureStatus(Enum):
    """Enum with all possible dynamic statuses."""

    NONE = "none"
    DYNAMIC_PALETTE = "dynamic_palette"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: str):
        """Set default enum member if an unknown value is provided."""
        return DynamicsFeatureStatus.UNKNOWN


@dataclass
class DynamicsFeature:
    """Represent DynamicsFeature object as used by the Hue api."""

    status: Optional[DynamicsFeatureStatus] = None
    status_values: Optional[List[DynamicsFeatureStatus]] = None
    # Duration of a light transition in ms. Accuracy is in 100ms steps.
    # minimal value 100, maximum 6000000
    duration: Optional[int] = None  # transitionspeed: only sent on update/set
    speed: Optional[float] = None  # speed for the dynamics
    speed_valid: Optional[bool] = None  # speed for the dynamics


class RecallAction(Enum):
    """Enum with available recall actions."""

    ACTIVE = "active"
    DYNAMIC_PALETTE = "dynamic_palette"


@dataclass
class RecallFeature:
    """
    Properties to send when updating/setting RecallFeature on the api.

    clip-api.schema.json#/definitions/RecallFeature
    """

    action: Optional[RecallAction] = None
    status: Optional[str] = None
    # Duration of transition in ms. Accuracy is in 100ms steps.
    # minimal value 100, maximum 6000000
    duration: Optional[int] = None
    dimming: Optional[DimmingFeatureBasic] = None


@dataclass
class PaletteColor:
    """Represents PaletteColor feature used for PaletteFeature."""

    color: Optional[ColorFeatureBasic] = None
    dimming: Optional[DimmingFeatureBasic] = None


@dataclass
class PaletteColorTemperature:
    """Represents PaletteColorTemperature feature used for PaletteFeature."""

    color_temperature: Optional[ColorTemperatureFeatureBasic] = None
    dimming: Optional[DimmingFeatureBasic] = None


@dataclass
class PaletteFeature:
    """Represents Palette feature used for dynamic scenes."""

    color: Optional[List[PaletteColor]] = None
    color_temperature: Optional[List[PaletteColorTemperature]] = None
    dimming: Optional[List[DimmingFeatureBasic]] = None



"""Model(s) for light_level resource on HUE bridge."""
from dataclasses import dataclass


from typing import Optional

from .resource import SensingService, ResourceTypes


@dataclass
class LightLevelFeature:
    """Represent LightLevelFeature."""

    light_level: Optional[int] = None
    light_level_valid: Optional[bool] = None


@dataclass
class LightLevel(SensingService):
    """
    Represent a LightLevel resource, a sensor reporting Illuminance in Lux.

    # TODO: CLIP Schema missing for this resource.
    """

    light: Optional[LightLevelFeature] = None
    type: ResourceTypes = ResourceTypes.LIGHT_LEVEL



"""Model(s) for bridge_home resource on HUE bridge."""
from dataclasses import dataclass

from .resource import ResourceTypes
from .group import Group


@dataclass
class BridgeHome(Group):
    """
    Represent BridgeHome object as retrieved from the api.

    Home resource lists all rooms in a home and all devices which are not assigned to a room.
    Home only contains resources of type "device" and "room"

    clip-api.schema.json#/definitions/BridgeHomeGet
    clip-api.schema.json#/definitions/BridgeHomePut
    """

    type: ResourceTypes = ResourceTypes.BRIDGE_HOME



"""Model(s) for script resource on HUE bridge."""
from dataclasses import dataclass


from typing import List, Optional

from .depender import ResourceDependeeGet
from .resource import Resource, ResourceMetadata, ResourceTypes


@dataclass
class ScriptMetadata(ResourceMetadata):
    """
    Represent ScriptMetadata object as used by the Hue api.

    clip-api.schema.json#/definitions/ScriptMetadataGet
    """

    name: Optional[str] = None


@dataclass
class ScriptDefinition(Resource):
    """
    Represent SceneService object as used by the Hue api.

    Definition of a script.
    clip-api.schema.json#/definitions/ScriptDefinitionGet
    """

    type: Optional[ResourceTypes]
    # Short description of script.
    description: Optional[str] = None
    # The configuration object is a JSON scheme that describes
    # the configuration needed to instantiate the script.
    configuration: Optional[dict] = None
    # version of the script
    version: Optional[str] = None
    metadata: Optional[ScriptMetadata] = None


@dataclass
class InstanceMetadata:
    """
    Represent InstanceMetadata object as used by the Hue api.

    Metadata for ScriptInstance.

    clip-api.schema.json#/definitions/InstanceMetadata
    """

    name: Optional[str] = None  # optional in get/post


@dataclass
@dataclass
class ScriptInstance(Resource):
    """
    Represent ScriptInstance object as used by the Hue api.

    Representation of active script.
    Scripts are instantiated using the id and configuration of the ScriptDefinition.

    clip-api.schema.json#/definitions/ScriptInstanceGet
    clip-api.schema.json#/definitions/ScriptInstancePost
    clip-api.schema.json#/definitions/ScriptInstancePut
    """

    # Identifier to ScriptDefinition.
    script_id: Optional[str] = None  # can not be set in post/put
    # Indicated whether a scripts is enabled.
    enabled: Optional[bool] = None
    # State of the instance according json-schema in corresponding
    # state attribute in the ScriptDefinition.
    state: Optional[dict] = None
    # Configuration of the script according json-schema in corresponding
    # config attribute in the ScriptDefinition.
    configuration: Optional[dict] = None
    # Represents all resources which this instance depends on.
    dependees: Optional[List[ResourceDependeeGet]] = None
    status: Optional[str] = None
    status_data: Optional[dict] = None
    metadata: Optional[InstanceMetadata] = None
    type: Optional[ResourceTypes] = None



"""Generic Group/Service Models used by various Hue resources."""

from dataclasses import dataclass
from typing import List, Optional, Set

from .resource import SENSOR_RESOURCE_TYPES, Resource, ResourceIdentifier, ResourceTypes


@dataclass
class Group(Resource):
    """
    Supertype of all groups grouping devices or services.

    clip-api.schema.json#/definitions/Group
    """

    children: Optional[
        List[ResourceIdentifier]
    ] = None  # only available in get response
    services: Optional[
        List[ResourceIdentifier]
    ] = None  # only available in get response
    # grouped_services references all services aggregating control and
    # state of services in the group. This includes all services grouped
    # in the group hierarchy given by child relation. This includes all services
    # of a device grouped in the group hierarchy given by child relation.
    # Aggregation is per service type, ie every service type which can be grouped
    # has a corresponding definition of grouped type.
    # Supported types: 'light'
    grouped_services: Optional[List[ResourceIdentifier]] = None

    @property
    def lights(self) -> Set[str]:
        """Return a set of light id's belonging to this group/device."""
        if not self.services:
            return set()
        return {x.rid for x in self.services if x.rtype == ResourceTypes.LIGHT}

    @property
    def sensors(self) -> Set[str]:
        """Return a set of sensor id's belonging to this group/device."""
        if not self.services:
            return set()
        return {x.rid for x in self.services if x.rtype in SENSOR_RESOURCE_TYPES}



"""Controller/model for Hue resource of type Config."""
from __future__ import annotations

from typing import Any, Coroutine, Dict


class Config:
    """
    Represent Hue config.

    https://developers.meethue.com/documentation/configuration-api#72_get_configuration
    """

    def __init__(self, raw: Dict[str, Any], request: Coroutine) -> None:
        """Initialize Config resource controller."""
        self.raw = raw
        self._request = request
        self.bridgeid = self.bridge_id  # for backwards compatability
        self.mac = self.mac_address  # for backwards compatability
        self.modelid = self.model_id  # for backwards compatability
        self.swversion = self.software_version  # for backwards compatability

    @property
    def bridge_id(self) -> str:
        """ID of the bridge."""
        return self.raw["bridgeid"]

    @property
    def name(self) -> str:
        """Name of the bridge."""
        return self.raw["name"]

    @property
    def mac_address(self) -> str:
        """Mac address of the bridge."""
        return self.raw["mac"]

    @property
    def model_id(self) -> str:
        """Model ID of the bridge."""
        return self.raw["modelid"]

    @property
    def software_version(self) -> str:
        """Software version of the bridge."""
        return self.raw["swversion"]

    @property
    def swupdate2_bridge_state(self) -> str:
        """Software update state of the bridge."""
        return self.raw.get("swupdate2", {}).get("bridge", {}).get("state")

    @property
    def apiversion(self) -> str:
        """Supported API version of the bridge."""
        return self.raw["apiversion"]

    async def update(self) -> None:
        """Update data for this resource."""
        self.raw = await self._request("get", "config")



"""Generic schemas for CLIP messages.."""
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, List, Type, TypeVar

from aiohue.util import dataclass_from_dict, parse_utc_timestamp

from ..models.light_level import LightLevel
from ..models.temperature import Temperature
from .behavior_instance import BehaviorInstance
from .behavior_script import BehaviorScript
from .bridge import Bridge
from .bridge_home import BridgeHome
from .button import Button
from .connectivity import ConnectivityService, ZigbeeConnectivity
from .depender import DependerGet
from .device import Device
from .device_power import DevicePower
from .entertainment import Entertainment, EntertainmentConfiguration
from .geofence_client import GeofenceClient
from .grouped_light import GroupedLight
from .homekit import HomekitGet
from .light import Light
from .motion import Motion
from .resource import Resource, ResourceTypes
from .room import Room
from .scene import Scene
from .zone import Zone

import logging

LOGGER = logging.getLogger(__package__)

CLIP_RESOURCE_MAPPING = {
    # Maps HUE resource type to correct class
    ResourceTypes.BEHAVIOR_INSTANCE: BehaviorInstance,
    ResourceTypes.BEHAVIOR_SCRIPT: BehaviorScript,
    ResourceTypes.BRIDGE: Bridge,
    ResourceTypes.BRIDGE_HOME: BridgeHome,
    ResourceTypes.BUTTON: Button,
    ResourceTypes.DEPENDER: DependerGet,
    ResourceTypes.DEVICE: Device,
    ResourceTypes.DEVICE_POWER: DevicePower,
    ResourceTypes.DEVICE_DISCOVERY: Resource,  # TODO
    ResourceTypes.DEVICE_UPDATE: Resource,  # TODO
    ResourceTypes.ENTERTAINMENT: Entertainment,
    ResourceTypes.ENTERTAINMENT_CONFIGURATION: EntertainmentConfiguration,
    ResourceTypes.GEOFENCE: Resource,  # TODO
    ResourceTypes.GEOFENCE_CLIENT: GeofenceClient,
    ResourceTypes.GROUPED_LIGHT: GroupedLight,
    ResourceTypes.HOMEKIT: HomekitGet,
    ResourceTypes.IP_CONNECTIVITY: ConnectivityService,  # TODO
    ResourceTypes.LIGHT: Light,
    ResourceTypes.LIGHT_LEVEL: LightLevel,
    ResourceTypes.MOTION: Motion,
    ResourceTypes.PUBLIC_IMAGE: Resource,  # TODO
    ResourceTypes.RELATIVE_ROTARY: Resource,  # TODO
    ResourceTypes.REMOTE_ACCESS: Resource,  # TODO
    ResourceTypes.ROOM: Room,
    ResourceTypes.SCENE: Scene,
    ResourceTypes.SYSTEM_UPDATE: Resource,  # TODO
    ResourceTypes.TEMPERATURE: Temperature,  # TODO
    ResourceTypes.ZIGBEE_BRIDGE_CONNECTIVITY: Resource,  # TODO
    ResourceTypes.ZIGBEE_CONNECTIVITY: ZigbeeConnectivity,
    ResourceTypes.ZONE: Zone,
    ResourceTypes.UNKNOWN: Resource,  # TODO
}
CLASS_TO_RESOURCE_MAPPING = {value: key for key, value in CLIP_RESOURCE_MAPPING.items()}

CLIPResource = TypeVar(
    "CLIPResource", *(Type[x] for x in CLIP_RESOURCE_MAPPING.values())
)


class CLIPEventType(Enum):
    """Type of the CLIP Event."""

    RESOURCE_ADDED = "add"
    RESOURCE_UPDATED = "update"
    RESOURCE_DELETED = "delete"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: Type, value: str):
        """Set default enum member if an unknown value is provided."""
        return CLIPEventType.UNKNOWN


def parse_clip_resource(obj_in: dict, strict=False) -> CLIPResource:
    """Parse raw object dict to correct CLIP Resource Class."""
    resource_type = ResourceTypes(obj_in["type"])
    resource_cls = CLIP_RESOURCE_MAPPING[resource_type]
    return dataclass_from_dict(resource_cls, obj_in, strict=False)


@dataclass
class CLIPEvent:
    """CLIP Event message as emitted by the EventStream."""

    id: str  # UUID
    creationtime: datetime
    type: CLIPEventType
    # data contains a list with resource objects
    # in case of add or update this is a full or partial resource object
    # in case of delete this will include only the
    # ResourceIndentifier (type and id) of the deleted object
    data: List[Resource]

    @classmethod
    def from_dict(cls: "CLIPEvent", dict_in: dict):
        """Create instance from dict."""
        return CLIPEvent(
            id=dict_in["id"],
            creationtime=parse_utc_timestamp(dict_in["creationtime"]),
            type=CLIPEventType(dict_in["type"]),
            data=[parse_clip_resource(x) for x in dict_in["data"]],
        )


@dataclass
class Error:
    """
    Error object which is returned when at least one error occured on a API call.

    clip-api.schema.json#/definitions/Error
    """

    # a human-readable explanation specific to this occurrence of the problem.
    description: str


@dataclass
class ClipMessage:
    """
    Basic message body for all return messages on GET, data contains full expanded resource.

    clip-api.schema.json#/definitions/ClipMessage
    """

    errors: List[Error]
    data: Any


@dataclass
class ClipMessageBehaviorInstance(ClipMessage):
    """
    CLIP Response when requesting behavior_instance resource.

    path: GET /clip/v2/resource/behavior_instance

    clip-api.schema.json#/definitions/ClipMessageBehaviorInstance
    """

    data: List[BehaviorInstance]


@dataclass
class ClipMessageBehaviorScript(ClipMessage):
    """
    CLIP Response when requesting behavior_script resource.

    path: GET /clip/v2/resource/behavior_script

    clip-api.schema.json#/definitions/ClipMessageBehaviorScript
    """

    data: List[BehaviorScript]


@dataclass
class ClipMessageBridge(ClipMessage):
    """
    CLIP Response when requesting bridge resource.

    path: GET /clip/v2/resource/bridge

    clip-api.schema.json#/definitions/ClipMessageBridge
    """

    data: List[Bridge]


@dataclass
class ClipMessageBridgeHome(ClipMessage):
    """
    CLIP Response when requesting bridge_home resource.

    path: GET /clip/v2/resource/bridge_home

    clip-api.schema.json#/definitions/ClipMessageBridgeHome
    """

    data: List[BridgeHome]


@dataclass
class ClipMessageButton(ClipMessage):
    """
    CLIP Response when requesting button resource.

    path: GET /clip/v2/resource/button

    clip-api.schema.json#/definitions/ClipMessageButton
    """

    data: List[Button]


@dataclass
class ClipMessageDependerGet(ClipMessage):
    """
    CLIP Response when requesting depender resource.

    path: GET /clip/v2/resource/depender

    clip-api.schema.json#/definitions/ClipMessageDependerGet
    """

    data: List[DependerGet]


@dataclass
class ClipMessageDevice(ClipMessage):
    """
    CLIP Response when requesting device resource.

    path: GET /clip/v2/resource/device

    clip-api.schema.json#/definitions/ClipMessageDevice
    """

    data: List[Device]


@dataclass
class ClipMessageEntertainmentConfiguration(ClipMessage):
    """
    CLIP Response when requesting entertainment_configuration resource.

    path: GET /clip/v2/resource/entertainment_configuration

    clip-api.schema.json#/definitions/ClipMessageEntertainmentConfiguration
    """

    data: List[EntertainmentConfiguration]


@dataclass
class ClipMessageEntertainment(ClipMessage):
    """
    CLIP Response when requesting entertainment resource.

    path: GET /clip/v2/resource/entertainment

    clip-api.schema.json#/definitions/ClipMessageEntertainment
    """

    data: List[Entertainment]


@dataclass
class ClipMessageGeofenceClient(ClipMessage):
    """
    CLIP Response when requesting geofence_client resource.

    path: GET /clip/v2/resource/geofence_client

    clip-api.schema.json#/definitions/ClipMessageGeofenceClient
    """

    data: List[GeofenceClient]


@dataclass
class ClipMessageGroupedLight(ClipMessage):
    """
    CLIP Response when requesting grouped_light resource.

    path: GET /clip/v2/resource/grouped_light

    clip-api.schema.json#/definitions/ClipMessageGroupedLight
    """

    data: List[GroupedLight]


@dataclass
class ClipMessageLight(ClipMessage):
    """
    CLIP Response when requesting light resource.

    path: GET /clip/v2/resource/light

    clip-api.schema.json#/definitions/ClipMessageLight
    """

    data: List[Light]


@dataclass
class ClipMessageMotion(ClipMessage):
    """
    CLIP Response when requesting motion resource.

    path: GET /clip/v2/resource/motion

    clip-api.schema.json#/definitions/ClipMessageMotion
    """

    data: List[Motion]


@dataclass
class ClipMessageResource(ClipMessage):
    """
    CLIP Response when requesting resource endpoint.

    path: GET /clip/v2/resource

    data will be an array containing all available resources inherited from Resource

    clip-api.schema.json#/definitions/ClipMessageResource
    """

    data: List[Resource]


@dataclass
class ClipMessageRoom(ClipMessage):
    """
    CLIP Response when requesting room resource.

    path: GET /clip/v2/resource/room

    clip-api.schema.json#/definitions/ClipMessageRoom
    """

    data: List[Room]


@dataclass
class ClipMessageScene(ClipMessage):
    """
    CLIP Response when requesting scene resource.

    path: GET /clip/v2/resource/scene

    clip-api.schema.json#/definitions/ClipMessageScene
    """

    data: List[Scene]


@dataclass
class ClipMessageZigbeeConnectivity(ClipMessage):
    """
    CLIP Response when requesting zigbee_connectivity resource.

    path: GET /clip/v2/resource/zigbee_connectivity

    clip-api.schema.json#/definitions/ClipMessageZigbeeConnectivity
    """

    data: List[ZigbeeConnectivity]


@dataclass
class ClipMessageZone(ClipMessage):
    """
    CLIP Response when requesting zone resource.

    path: GET /clip/v2/resource/zone

    clip-api.schema.json#/definitions/ClipMessageZone
    """

    data: List[Zone]



"""Setup for aiohue."""
from setuptools import find_packages, setup

LONG_DESC = open("README.md").read()
PACKAGES = find_packages(exclude=["tests", "tests.*"])
REQUIREMENTS = list(val.strip() for val in open("requirements.txt"))
MIN_PY_VERSION = "3.8"

setup(
    name="aiohue",
    version="3.0.10",
    license="Apache License 2.0",
    url="https://github.com/home-assistant-libs/aiohue",
    author="Paulus Schoutsen",
    author_email="paulus@paulusschoutsen.nl",
    description="Python module to talk to Philips Hue.",
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    packages=PACKAGES,
    zip_safe=True,
    platforms="any",
    install_requires=REQUIREMENTS,
    python_requires=f">={MIN_PY_VERSION}",
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



"""Model(s) for device resource on HUE bridge."""
from dataclasses import dataclass
from enum import Enum

from typing import Optional, Type

from .group import Group
from .resource import NamedResourceMetadata, ResourceTypes


class DeviceArchetypes(Enum):
    """
    Enum with all possible Device archtypes.

    clip-api.schema.json#/definitions/DeviceArchetypes
    """

    BRIDGE_V2 = "bridge_v2"
    UNKNOWN_ARCHETYPE = "unknown_archetype"
    CLASSIC_BULB = "classic_bulb"
    SULTAN_BULB = "sultan_bulb"
    FLOOD_BULB = "flood_bulb"
    SPOT_BULB = "spot_bulb"
    CANDLE_BULB = "candle_bulb"
    LUSTER_BULB = "luster_bulb"
    PENDANT_ROUND = "pendant_round"
    PENDANT_LONG = "pendant_long"
    CEILING_ROUND = "ceiling_round"
    CEILING_SQUARE = "ceiling_square"
    FLOOR_SHADE = "floor_shade"
    FLOOR_LANTERN = "floor_lantern"
    TABLE_SHADE = "table_shade"
    RECESSED_CEILING = "recessed_ceiling"
    RECESSED_FLOOR = "recessed_floor"
    SINGLE_SPOT = "single_spot"
    DOUBLE_SPOT = "double_spot"
    TABLE_WASH = "table_wash"
    WALL_LANTERN = "wall_lantern"
    WALL_SHADE = "wall_shade"
    FLEXIBLE_LAMP = "flexible_lamp"
    GROUND_SPOT = "ground_spot"
    WALL_SPOT = "wall_spot"
    PLUG = "plug"
    HUE_GO = "hue_go"
    HUE_LIGHTSTRIP = "hue_lightstrip"
    HUE_IRIS = "hue_iris"
    HUE_BLOOM = "hue_bloom"
    BOLLARD = "bollard"
    WALL_WASHER = "wall_washer"
    HUE_PLAY = "hue_play"
    VINTAGE_BULB = "vintage_bulb"
    CHRISTMAS_TREE = "christmas_tree"
    HUE_CENTRIS = "hue_centris"
    HUE_LIGHTSTRIP_TV = "hue_lightstrip_tv"

    @classmethod
    def _missing_(cls: Type, value: str):
        """Set default enum member if an unknown value is provided."""
        return DeviceArchetypes.UNKNOWN_ARCHETYPE


@dataclass
class DeviceProductData:
    """
    Represent a DeviceProductData object as used by the Hue api.

    clip-api.schema.json#/definitions/DeviceProductData
    """

    model_id: str
    manufacturer_name: str
    product_name: str
    product_archetype: DeviceArchetypes
    certified: bool
    software_version: str
    product_id: Optional[str] = None  # missing in output from bridge ?


@dataclass
class DeviceMetaData(NamedResourceMetadata):
    """
    Represent DeviceMetaData object as used by the Hue api.

    clip-api.schema.json#/definitions/DeviceMetaData
    """

    archetype: Optional[DeviceArchetypes] = None


@dataclass
class Device(Group):
    """
    Represent a Device object as used by the Hue api.

    clip-api.schema.json#/definitions/DeviceGet
    clip-api.schema.json#/definitions/DevicePut
    """

    product_data: Optional[DeviceProductData] = None
    metadata: Optional[DeviceMetaData] = None  # optional in put
    creation_time: Optional[str] = None
    type: ResourceTypes = ResourceTypes.DEVICE



"""Model(s) for behavior_script resource on HUE bridge."""
from dataclasses import dataclass

from .resource import ResourceTypes
from .script import ScriptDefinition


@dataclass
class BehaviorScript(ScriptDefinition):
    """
    Represent BehaviorScript object as used by the Hue api.

    clip-api.schema.json#/definitions/BehaviorScriptGet
    """

    type: ResourceTypes = ResourceTypes.BEHAVIOR_SCRIPT



"""Dependency Schemas used by (script) Hue resources."""

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

from .resource import Resource, ResourceIdentifier, ResourceTypes


class DependencyLevel(Enum):
    """
    Enum with dependency Levels.

    - non_critical: Defines a dependency between resources:
      although source is impacted by removal of target, source remains
      of target means removal of source.
    - critical: Defines a critical dependency between resources:
      source cannot function without target,
      hence removal of target means removal of source.
    """

    NON_CRITICAL = "non_critical"
    CRITICAL = "critical"


@dataclass
class DependencyGet:
    """
    Dependency object as used by the Hue api.

    Defines target resource of a dependency.

    clip-api.schema.json#/definitions/DependencyGet
    """

    target: ResourceIdentifier  # Id of the dependency resource (target).
    level: DependencyLevel
    type: Optional[str] = None


@dataclass
class ResourceDependeeGet(DependencyGet):
    """
    ResourceDependee object as used by the Hue api.

    Represents a resource which (this) resource is dependent on.

    clip-api.schema.json#/definitions/ResourceDependeeGet
    """


@dataclass
class ResourceDependerGet(DependencyGet):
    """
    ResourceDepender object as used by the Hue api.

    Represents a resource which depend on (this) resource.

    clip-api.schema.json#/definitions/ResourceDependerGet
    """


@dataclass
class DependerGet(Resource):
    """
    Depender object as used by the Hue api.

    Represents all the resources which depend on a resource.

    clip-api.schema.json#/definitions/DependerGet
    """

    dependers: List[ResourceDependerGet] = field(default_factory=list)
    type: ResourceTypes = ResourceTypes.DEPENDER



"""Model(s) for homekit resource on HUE bridge."""
from dataclasses import dataclass, field
from enum import Enum

from typing import List, Optional

from .resource import Resource, ResourceTypes


class HomekitStatus(Enum):
    """Enum with the possible homekit status."""

    PAIRING = "pairing"
    PAIRED = "paired"
    UNPAIRED = "unpaired"


@dataclass
class HomekitGet(Resource):
    """Represent a Homekit resource object as used by the Hue api."""

    status: Optional[HomekitStatus] = None
    status_values: List[HomekitStatus] = field(
        default_factory=HomekitStatus.__members__.values
    )
    type: ResourceTypes = ResourceTypes.HOMEKIT



"""Model(s) for device_power resource on HUE bridge."""
from dataclasses import dataclass
from enum import Enum


from typing import Optional, Type

from .resource import Resource, ResourceTypes


class BatteryState(Enum):
    """
    Enum with all possible BatteryStates.

    clip-api.schema.json#/definitions/TODO
    """

    UNKNOWN = "unknown"
    NORMAL = "normal"
    LOW = "low"

    @classmethod
    def _missing_(cls: Type, value: str):
        """Set default enum member if an unknown value is provided."""
        return BatteryState.UNKNOWN


@dataclass
class PowerState:
    """
    Represent PowerState as retrieved from api.

    clip-api.schema.json#/definitions/TODO
    """

    battery_level: Optional[int]
    battery_state: BatteryState


@dataclass
class DevicePower(Resource):
    """
    Represent a DevicePower resource, a sensor with battery state.

    clip-api.schema.json#/definitions/TODO
    """

    power_state: Optional[PowerState] = None
    type: ResourceTypes = ResourceTypes.DEVICE_POWER



"""Model(s) for zigbee_connectivity resource on HUE bridge."""
from dataclasses import dataclass
from enum import Enum


from typing import Optional

from .resource import Resource, ResourceTypes


class ConnectivityServiceStatus(Enum):
    """
    Enum with possible ConnectivityService statuses.

    defined in clip-api.schema.json#/definitions/ConnectivityService
    """

    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    CONNECTIVITY_ISSUE = "connectivity_issue"


@dataclass
class ConnectivityService(Resource):
    """
    Represent ConnectivityService object as used by the Hue api.

    clip-api.schema.json#/definitions/ConnectivityService
    """

    status: Optional[ConnectivityServiceStatus] = None  # required in get


@dataclass
class ZigbeeConnectivity(ConnectivityService):
    """
    Represent ZigbeeConnectivity object as used by the Hue api.

    clip-api.schema.json#/definitions/ZigbeeConnectivity
    """

    mac_address: Optional[str] = None  # can not be set
    type: ResourceTypes = ResourceTypes.ZIGBEE_CONNECTIVITY
